import sys
import itertools
import collections
import pygame
import ptext

import data
import level
import const
import util
import story

class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Breach')
        # self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT), pygame.FULLSCREEN)
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath("floating_cities.mp3"))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load sound. :('
        self.intro()
    
    def intro(self):
        Intro(self, self.screen).loop()
        self.game()
    
    def game(self):
        nlevels = 12
        for levelnum in range(1, nlevels + 1):
            restart = True
            while restart:
                game = Game(self, self.screen, levelnum)
                game.loop()
                restart = game.restarted
            story_id = levelnum - 1
            story.Story(self, self.screen, story_id).loop()
            if story_id == len(story.STORY) - 1:
                story.Story(self, self.screen, story_id + 1).loop()
                break          
        story.Outro(self, self.screen).loop()


class Intro(object):

    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.bg = pygame.image.load(data.filepath("intro.png")).convert()
        self.screen.blit(self.bg, (0, 0))

        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()
        
    def loop(self):
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
                    else:
                        self.done = True


class Game(object):

    def __init__(self, window, screen, levelnum):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.clock = pygame.time.Clock()

        self.paused = False

        self.levelnum = levelnum
        self.level = level.LevelManager(data.filepath('level' + str(self.levelnum) + '.tmx'))
        self.character = Character(self.level.get_start_pos())
        self.character_con = pygame.sprite.RenderUpdates(self.character)
        self.doors = DoorManager(self.level.get_layer("Doors"), self.level.tmx_data)
        self.slots = SlotManager(self.level.get_layer("Slots"))
        self.operator_front = pygame.sprite.RenderUpdates()
        self.operators = OperatorManager(self.level.get_layer("Operators"), self.operator_front, self.slots)
        self.wires = WireManager(self.level.get_layer("Wires"), self.level.tmx_data, self.slots)
        self.sources = SourceManager(self.level.get_layer("Sources"), self.level.tmx_data, self.wires, self.doors)
        self.tracks = TrackManager(self.level.get_layer("Tracks"), self.level.tmx_data)
        self.breaches = BreachManager(self.level.get_layer("Breaches"), self.level.tmx_data, self.tracks, self.doors, self.sources)
        self.textlayer = TextLayer(self.levelnum)
        self.textlayer.draw_level_title()

        self.update_level_state()

        self.col_rects = self.level.get_inner_wall_rects()
 
        self.collision_mgr = CollisionManager()

        self.restarted = False
        self.lost = False
        self.won = False

    def loop(self):
        self.bg = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.level.render_map(self.bg)
        self.screen.blit(self.bg, (0, 0))

        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()

        self.done = False

        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)

            if not self.paused:
                self.tracks.clear(self.screen, self.bg)
                self.wires.clear(self.screen, self.bg)
                self.operators.clear(self.screen, self.bg)
                self.doors.clear(self.screen, self.bg)
                self.breaches.clear(self.screen, self.bg)
                self.sources.clear(self.screen, self.bg)
                self.slots.clear(self.screen, self.bg)
                self.character_con.clear(self.screen, self.bg)
                self.operator_front.clear(self.screen, self.bg)
                self.textlayer.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0

            if not self.paused:
                self.tracks.update(dt)
                self.wires.update(dt)
                self.operators.update(dt)
                self.doors.update(dt)
                self.breaches.update(dt)
                self.sources.update(dt)
                self.slots.update(dt)
                self.character_con.update(dt)
                self.collision_mgr.handle_collisions(self.character, self.level, self.doors, self.col_rects)
                self.operator_front.update(dt)
                self.textlayer.update(dt)

                self.update(dt)

            update_rects = []
            if not self.paused:
                update_rects += self.tracks.draw(self.screen)
                update_rects += self.wires.draw(self.screen)
                update_rects += self.operators.draw(self.screen)
                update_rects += self.doors.draw(self.screen)
                update_rects += self.breaches.draw(self.screen)
                update_rects += self.sources.draw(self.screen)
                update_rects += self.slots.draw(self.screen)
                update_rects += self.character_con.draw(self.screen)
                update_rects += self.operator_front.draw(self.screen)
                update_rects += self.textlayer.draw(self.screen)

            scaled_update_rects = [pygame.Rect(2*r.left, 2*r.top, 2*r.width, 2*r.height) for r in update_rects]

            pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
            pygame.display.update(scaled_update_rects)
            # print self.clock.get_fps()

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            if self.paused:
                sys.exit()
            else:
                self.pause()
        elif event.key == pygame.K_p:
            self.pause()
        if not self.lost:
            if event.key == pygame.K_UP or event.key == pygame.K_w:
                self.character.move_up()
            elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                self.character.move_down()
            elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
                self.character.move_left()
            elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
                self.character.move_right()

    def on_keyup(self, event):
        if not self.lost:
            if event.key == pygame.K_UP or event.key == pygame.K_w:
                self.character.stop_up()
            elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                self.character.stop_down()
            elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
                self.character.stop_left()
            elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
                self.character.stop_right()
            elif event.key in [pygame.K_e, pygame.K_z, pygame.K_x, pygame.K_SPACE, pygame.K_j]:
                self.character.interact(self.operators, self.slots, self.sources, self.breaches, self.doors, self.win, self.update_level_state)
        if event.key == pygame.K_r:
            self.restart()
        elif event.key == pygame.K_n:
            self.win()

    def pause(self):
        self.paused = not self.paused

    def restart(self):
        self.done = True
        self.restarted = True

    def lose(self):
        self.lost = True
        self.textlayer.draw_lose_text()
        sound = pygame.mixer.Sound(data.filepath("lose.ogg"))
        sound.set_volume(const.SOUND_VOLUME)
        sound.play()

    def win(self):
        self.won = True
        self.done = True

    def update(self, dt):
        if not self.lost and self.breaches.breached:
            self.lose()

    def update_level_state(self):
        self.sources.update_wires()
        source_updated = self.breaches.update_tracks()
        niter = 0
        # Repeat for track-powered sources; note that levels cannot have circular dependency
        while source_updated:
            self.sources.update_wires()
            source_updated = self.breaches.update_tracks()
            niter += 1
            if niter >= 5:
                break # probably circular dependency, give up
        self.breaches.apply_breached_tracks() # breaching is postponed


class Character(pygame.sprite.Sprite):

    DOWN = 0
    UP = 1
    LEFT = 2
    RIGHT = 3

    def __init__(self, pos):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.last_pos = list(pos)
        self.vel = [0, 0]

        self.image = pygame.image.load(data.filepath('character.png')).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.bb = pygame.Rect(8, 11, 16, 19)
        self.in_hand = None

        self.moving_left = False
        self.moving_right = False
        self.moving_up = False
        self.moving_down = False

        self.speed = 100

    def update(self, dt):
        self.update_move()
        self.last_pos[0] = self.pos[0]
        self.last_pos[1] = self.pos[1]
        self.pos[0] += self.vel[0] * dt
        self.pos[1] += self.vel[1] * dt
        self.rect.topleft = self.pos

    def move_left(self):  self.moving_left = True
    def stop_left(self):  self.moving_left = False
    def move_right(self): self.moving_right = True
    def stop_right(self): self.moving_right = False
    def move_up(self):    self.moving_up = True
    def stop_up(self):    self.moving_up = False
    def move_down(self):  self.moving_down = True
    def stop_down(self):  self.moving_down = False

    def update_move(self):
        if self.moving_left: self.vel[0] = -self.speed
        elif self.moving_right: self.vel[0] = self.speed
        else: self.vel[0] = 0
        if self.moving_up: self.vel[1] = -self.speed
        elif self.moving_down: self.vel[1] = self.speed
        else: self.vel[1] = 0

    def move_to(self, x, y):
        self.pos[0] = x
        self.pos[1] = y
        self.rect.topleft = self.pos

    def interact(self, operators, slots, sources, breaches, doors, win, update_level_state):
        tilepos = util.get_tile_at_xy(*self.rect.center)
        closed = breaches.close_adjacent_breaches(*tilepos)
        if closed:
            update_level_state()
        if breaches.n_open == 0:
            # done with level
            doors.force_all_open()
            win()
        if self.in_hand == None:
            operator = operators.get_operator_at_xy(*self.rect.center)
            if operator:
                prev_opslot = operator.slot
                self.pick_up(operator)
                if prev_opslot != None:
                    update_level_state()
        else:
            slot = slots.slot_map.get(tilepos)
            self.drop(slot, tilepos)
            if slot != None:
                update_level_state()

    def pick_up(self, operator):
        operator.pick_up(self)
        self.in_hand = operator

    def drop(self, slot, tilepos):
        can_drop = self.in_hand.drop(slot, tilepos)
        if can_drop:
            self.in_hand = None


class CollisionManager(object):
    
    def handle_collisions(self, character, level, doors, col_rects):
        self.handle_sprite_tilemap_collision(character, level)
        door_rects = [pygame.Rect(door.rect.x + door.bb.x, door.rect.y + door.bb.y, door.bb.width, door.bb.height)
            for door in doors if door.closed and not door.oneway]
        self.handle_sprite_rects_collision(character, door_rects)
        onewaydoor_rects = [(pygame.Rect(door.rect.x + door.bb.x, door.rect.y + door.bb.y, door.bb.width, door.bb.height), door.oneway)
            for door in doors if door.closed and door.oneway]
        self.handle_sprite_rects_collision_single_dir(character, onewaydoor_rects)
        self.handle_sprite_rects_collision(character, col_rects)
    
    def handle_sprite_tilemap_collision(self, sprite, level):
        offset = 0.5
        bb_left = sprite.bb.left
        bb_right = sprite.rect.width - sprite.bb.right
        bb_top = sprite.bb.top
        bb_bottom = sprite.rect.height - sprite.bb.bottom
        old_left = sprite.last_pos[0] + bb_left + offset
        old_right = sprite.last_pos[0] + sprite.rect.width - bb_right - offset
        old_top = sprite.last_pos[1] + bb_top + offset
        old_bottom = sprite.last_pos[1] + sprite.rect.height - bb_bottom - offset
        x = sprite.pos[0]
        y = sprite.pos[1]
        dx = sprite.pos[0] - sprite.last_pos[0] 
        dy = sprite.pos[1] - sprite.last_pos[1]
        
        if dx > 0:
            right = sprite.pos[0] + sprite.rect.width - bb_right - offset
            top_right = util.get_tile_at_xy(right, old_top)
            bottom_right = util.get_tile_at_xy(right, old_bottom)
            if level.is_solid(*top_right) or level.is_solid(*bottom_right):
                x = (top_right[0] - 1) * const.TILE_SIZE + bb_right
        elif dx < 0:
            left = sprite.pos[0] + bb_left + offset
            top_left = util.get_tile_at_xy(left, old_top)
            bottom_left = util.get_tile_at_xy(left, old_bottom)
            if level.is_solid(*top_left) or level.is_solid(*bottom_left):
                x = (top_left[0] + 1) * const.TILE_SIZE - bb_left

        if dy > 0:
            bottom = sprite.pos[1] + sprite.rect.height - bb_bottom - offset
            bottom_left = util.get_tile_at_xy(old_left, bottom)
            bottom_right = util.get_tile_at_xy(old_right, bottom)
            if level.is_solid(*bottom_left) or level.is_solid(*bottom_right):
                y = (bottom_left[1] - 1) * const.TILE_SIZE + bb_bottom
        elif dy < 0:
            top = sprite.pos[1] + bb_top + offset
            top_left = util.get_tile_at_xy(old_left, top)
            top_right = util.get_tile_at_xy(old_right, top)
            if level.is_solid(*top_left) or level.is_solid(*top_right):
                y = (top_left[1] + 1) * const.TILE_SIZE - bb_top
        
        if x != sprite.rect.x or y != sprite.rect.y:
            sprite.move_to(x, y)

    def handle_sprite_rects_collision(self, sprite, static_rects):
        for srect in static_rects:
            x = sprite.rect.x
            sprite_rect_x = pygame.Rect(x + sprite.bb.x, sprite.last_pos[1] + sprite.bb.y, sprite.bb.width, sprite.bb.height)
            if sprite_rect_x.colliderect(srect):
                dx = sprite.pos[0] - sprite.last_pos[0]
                if dx > 0:
                    x = srect.left - sprite.bb.x - sprite.bb.width
                elif dx < 0:
                    x = srect.right - sprite.bb.x
            if x != sprite.rect.x:
                sprite.move_to(x, sprite.pos[1])
        for srect in static_rects:
            x = sprite.rect.x
            y = sprite.rect.y
            sprite_rect_y = pygame.Rect(x + sprite.bb.x, y + sprite.bb.y, sprite.bb.width, sprite.bb.height)
            if sprite_rect_y.colliderect(srect):
                dy = sprite.pos[1] - sprite.last_pos[1]
                if dy > 0:
                    y = srect.top - sprite.bb.y - sprite.bb.height
                elif dy < 0:
                    y = srect.bottom - sprite.bb.y
            if y != sprite.rect.y:
                sprite.move_to(sprite.pos[0], y)

    def handle_sprite_rects_collision_single_dir(self, sprite, static_rects_dirs):
        for srect, direction in static_rects_dirs:
            dir_to_block = util.opposite_dir(direction)
            x = sprite.rect.x
            y = sprite.rect.y
            if dir_to_block == 'l' or dir_to_block == 'r':
                sprite_rect_x = pygame.Rect(x + sprite.bb.x, y + sprite.bb.y, sprite.bb.width, sprite.bb.height)
                if sprite_rect_x.colliderect(srect):
                    dx = sprite.pos[0] - sprite.last_pos[0]
                    if dir_to_block == 'r' and dx > 0: # right
                        x = srect.left - sprite.bb.x - sprite.bb.width
                    elif dir_to_block == 'l' and dx < 0: # left
                        x = srect.right - sprite.bb.x
                if x != sprite.rect.x:
                    sprite.move_to(x, sprite.pos[1])
            elif dir_to_block == 't' or dir_to_block == 'b':
                dir_to_block = util.opposite_dir(direction)
                sprite_rect_y = pygame.Rect(x + sprite.bb.x, y + sprite.bb.y, sprite.bb.width, sprite.bb.height)
                if sprite_rect_y.colliderect(srect):
                    dy = sprite.pos[1] - sprite.last_pos[1]
                    if dir_to_block == 'b' and dy > 0: # bottom
                        y = srect.top - sprite.bb.y - sprite.bb.height
                    elif dir_to_block == 't' and dy < 0: # top
                        y = srect.bottom - sprite.bb.y
                if y != sprite.rect.y:
                    sprite.move_to(sprite.pos[0], y)


class OperatorManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_operators, front, slots):
        pygame.sprite.RenderUpdates.__init__(self)
        self.operator_map = {}
        self.front = front # renderer that is in front of character; for picked up objects
        for tmx_operator in tmx_operators:
            if 'optype' not in tmx_operator.properties and tmx_operator.properties['optype'] not in ["not", "and", "or", "xor"]:
                print "Warning: Unknown operator in level ignored"
                continue
            tilepos = util.get_tile_at_xy(tmx_operator.x, tmx_operator.y)
            slot = slots.slot_map.get(tilepos)
            op = Operator(tmx_operator, self, slot)
            self.add(op)
            self.operator_map[tilepos] = op

    def get_operator_at_xy(self, x, y):
        tilepos = util.get_tile_at_xy(x, y)
        if tilepos not in self.operator_map:
            return None
        return self.operator_map[tilepos]

    def operator_map_remove(self, operator):
        keys_to_remove = [k for k, v in self.operator_map.iteritems() if v == operator]
        for k in keys_to_remove:
            self.operator_map.pop(k)

    def operator_map_add(self, operator, tilepos=None):
        if not tilepos:
            tilepos = util.get_tile_at_xy(*operator.rect.center)
        if tilepos in self.operator_map:
            return False
        self.operator_map[tilepos] = operator
        return True


class Operator(pygame.sprite.Sprite):

    def __init__(self, tmx_operator, mgr, slot):
        pygame.sprite.Sprite.__init__(self)
        self.type_op = tmx_operator.properties['optype']
        self.offset = 5
        self.pos = [tmx_operator.x + self.offset, tmx_operator.y + self.offset]
        self.mgr = mgr

        self.image = tmx_operator.image
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos

        self.pick_up_sound = pygame.mixer.Sound(data.filepath("get_slot.ogg"))
        self.pick_up_sound.set_volume(const.SOUND_VOLUME)
        self.put_sound = pygame.mixer.Sound(data.filepath("put_slot.ogg"))
        self.put_sound.set_volume(const.SOUND_VOLUME)

        self.picked_up = False
        if slot:
            self.pos = [tmx_operator.x, tmx_operator.y]
            self.slot = slot
            slot.attach_operator(self)
        else:
            self.slot = None

    def update(self, dt):
        if self.picked_up:
            self.pos[0] = self.character.pos[0] + self.offset
            self.pos[1] = self.character.pos[1] + self.offset
        self.rect.topleft = self.pos
        # print self.type_op, self.pos

    def pick_up(self, character):
        self.picked_up = True
        self.character = character
        self.mgr.operator_map_remove(self)
        self.mgr.remove(self)
        self.mgr.front.add(self)
        if self.slot:
            self.pick_up_sound.play()
            self.slot.detach_operator()
            self.slot = None

    def drop(self, slot, tilepos):
        can_drop = self.mgr.operator_map_add(self, tilepos)
        if not can_drop: return False
        self.picked_up = False
        self.pos = list(util.get_xy_at_tile(*tilepos))
        if slot != None:
            self.put_sound.play()
            self.slot = slot
            slot.attach_operator(self)
        else:
            self.pos[0] += self.offset
            self.pos[1] += self.offset
        self.mgr.front.remove(self)
        self.mgr.add(self)
        return True


class DoorManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_doors, tmx_data):
        pygame.sprite.RenderUpdates.__init__(self)
        self.door_map = collections.defaultdict(list)
        for tmx_door in tmx_doors:
            door = Door(tmx_door, tmx_data)
            self.add(door)
            door_tile = util.get_tile_at_xy(tmx_door.x, tmx_door.y)
            self.door_map[door_tile].append(door)
        for door in self:
            door.create_siblings(self.door_map) # link wide doors together

    def force_all_open(self):
        for door in self:
            door.force_open()


class Door(pygame.sprite.Sprite):

    DOOR_WIDTH = 6

    def __init__(self, tmx_door, tmx_data):
        pygame.sprite.Sprite.__init__(self)
        self.direction = tmx_door.properties["doordir"]
        self.pos = [tmx_door.x, tmx_door.y]
        self.tilepos = util.get_tile_at_xy(self.pos[0], self.pos[1])
        self.oneway = tmx_door.properties.get("oneway")

        self.image_on = tmx_door.image
        self.image_off = tmx_data.get_tile_image_by_gid(tmx_door.properties['frames'][1].gid)
        self.image = self.image_off
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.siblings = []
        self.closed = False
        self.forced_open = False

        if self.direction == "top":
            self.bb = pygame.Rect(0, 0, self.rect.width, self.DOOR_WIDTH)
        elif self.direction == "bottom":
            self.bb = pygame.Rect(0, self.rect.height - self.DOOR_WIDTH, self.rect.width, self.DOOR_WIDTH)
        elif self.direction == "left":
            self.bb = pygame.Rect(0, 0, self.DOOR_WIDTH, self.rect.height)
        elif self.direction == "right":
            self.bb = pygame.Rect(self.rect.width - self.DOOR_WIDTH, 0, self.DOOR_WIDTH, self.rect.height)

    def create_siblings(self, door_map):
        # siblings are door tiles part of a wide door; only neighboring doors
        self.siblings = []
        sides = {"top": ['l', 'r'], "bottom": ['l', 'r'], "left": ['t', 'b'], "right": ['t', 'b']}
        for side_dir in sides[self.direction]:
            tpos = util.get_tile_from_dir(self.tilepos[0], self.tilepos[1], side_dir)
            sibling = next((door for door in door_map[tpos] if door.direction == self.direction), None)
            if sibling: self.siblings.append(sibling)

    def close(self):
        if not self.closed and not self.forced_open:
            self.closed = True
            self.image = self.image_on
            for sibling in self.siblings:
                sibling.close()

    def open(self):
        if self.closed:
            self.closed = False
            self.image = self.image_off
            for sibling in self.siblings:
                sibling.open()

    def force_open(self):
        self.open()
        self.forced_open = True


class BreachManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_breaches, tmx_data, tracks, doors, sources):
        pygame.sprite.RenderUpdates.__init__(self)
        self.breach_map = {}
        self.n_open = 0
        for tmx_breach in tmx_breaches:
            breach = Breach(tmx_breach, tmx_data, tracks, doors, sources)
            self.add(breach)
            self.n_open += 1
            breach_tile = util.get_tile_at_xy(tmx_breach.x, tmx_breach.y)
            if breach_tile in self.breach_map:
                print "Error loading level: Overlapping breaches"
                sys.exit()
            self.breach_map[breach_tile] = breach
        self.breached = False

    def update_tracks(self):
        update_source = False
        for breach in self:
            blocked, breach_update_source = breach.update_tracks()
            if breach_update_source: update_source = True
            # if not blocked:
                # breach.animate_tracks()
                # self.breached = True
        return update_source

    def apply_breached_tracks(self):
        for breach in self:
            blocked, breach_update_source = breach.update_tracks()
            if not blocked:
                breach.animate_tracks()
                self.breached = True

    def get_adjacent_breaches(self, tx, ty):
        breaches = []
        for direction in util.DIRS:
            tp = util.get_tile_from_dir(tx, ty, direction)
            if tp != None and tp in self.breach_map:
                breach = self.breach_map[tp]
                if breach.direction == util.opposite_dir(direction):
                    breaches.append(breach)
        return breaches

    def close_adjacent_breaches(self, tx, ty):
        closed_any = False
        for breach in self.get_adjacent_breaches(tx, ty):
            closed = breach.close()
            if closed:
                self.n_open -= 1
                closed_any = True
        return closed_any


class Breach(pygame.sprite.Sprite):

    def __init__(self, tmx_breach, tmx_data, all_tracks, doors, sources):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [tmx_breach.x, tmx_breach.y]

        self.direction = tmx_breach.properties['dir']
        self.image_on = tmx_breach.image
        self.image_off = tmx_data.get_tile_image_by_gid(tmx_breach.properties['frames'][1].gid)
        self.image = self.image_on
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.sources = sources

        self.closed = False

        self.extract_tracks(tmx_breach, all_tracks, doors)

    def extract_tracks(self, tmx_breach, all_tracks, doors):
        self.tracks_and_doors = []
        next_pos = util.get_tile_at_xy(tmx_breach.x, tmx_breach.y)
        next_dir = self.direction
        while next_pos != None:
            next_pos = util.get_tile_from_dir(next_pos[0], next_pos[1], next_dir)
            if next_pos == None: break
            track = all_tracks.track_map[next_pos]

            # door blocking 'in' direction
            door_in_way = next((door for door in doors.door_map[next_pos] if door.direction[0] == util.opposite_dir(track.out_dir)
                and door.oneway != track.out_dir), None)
            if door_in_way: self.tracks_and_doors.append(door_in_way)

            # track
            next_dir = track.out_dir
            self.tracks_and_doors.append(track)

            # door blocking 'out' direction
            door_in_way = next((door for door in doors.door_map[next_pos] if door.direction[0] == track.out_dir
                and door.oneway != track.out_dir), None)
            if door_in_way: self.tracks_and_doors.append(door_in_way)

    def update_tracks(self):
        # return False if reached end of track
        door_blocked = self.closed
        updated_center_source = False
        for track_or_door in self.tracks_and_doors:
            if isinstance(track_or_door,Track):
                if door_blocked:
                    track_or_door.deactivate()
                else:
                    track_or_door.activate()

                # check for center sources
                if track_or_door.tilepos in self.sources.source_map:
                    source = self.sources.source_map[track_or_door.tilepos]
                    if source.direction == 'c':
                        if door_blocked:
                            if source.on == True:
                                source.disable()
                                updated_center_source = True
                        else:
                            if source.on == False:
                                source.enable()
                                updated_center_source = True
            else:
                if track_or_door.closed:
                    door_blocked = True
        return (door_blocked, updated_center_source)

    def animate_tracks(self):
        st = 0
        for track in self.tracks_and_doors:
            if isinstance(track, Track):
                track.anim.set_start_time(st)
                track.animate()
                st += 0.1

    def close(self):
        if not self.closed:
            self.closed = True
            self.image = self.image_off
            self.update_tracks()
            sound = pygame.mixer.Sound(data.filepath("close_breach.ogg"))
            sound.set_volume(const.SOUND_VOLUME)
            sound.play()
            return True
        return False


class SlotManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_slots):
        pygame.sprite.RenderUpdates.__init__(self)
        self.slot_map = {}
        for tmx_slot in tmx_slots:
            slot = Slot(tmx_slot)
            self.add(slot)
            slot_tile = util.get_tile_at_xy(tmx_slot.x, tmx_slot.y)
            if slot_tile in self.slot_map:
                print "Error loading level: Overlapping slots"
                sys.exit()
            self.slot_map[slot_tile] = slot


class Slot(pygame.sprite.Sprite):

    def __init__(self, tmx_slot):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [tmx_slot.x, tmx_slot.y]

        self.image = tmx_slot.image
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos

        self.not_gate = False
        self.operator = None

    def attach_operator(self, operator):
        self.operator = operator
        self.not_gate = True

    def detach_operator(self):
        self.operator = None
        self.not_gate = False


class TrackManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_tracks, tmx_data):
        pygame.sprite.RenderUpdates.__init__(self)
        self.track_map = {}
        for tmx_track in tmx_tracks:
            track = Track(tmx_track, tmx_data)
            self.add(track)
            track_tile = util.get_tile_at_xy(tmx_track.x, tmx_track.y)
            if track_tile in self.track_map:
                print "Error loading level: Overlapping tracks"
                sys.exit()
            self.track_map[track_tile] = track


class Track(pygame.sprite.Sprite):

    # Activated tracks are found in the track's animation

    def __init__(self, tmx_track, tmx_data):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [tmx_track.x, tmx_track.y]
        self.tilepos = util.get_tile_at_xy(*self.pos)

        self.out_dir = tmx_track.properties['out']
        self.image_off = tmx_track.image
        self.image_on = tmx_data.get_tile_image_by_gid(tmx_track.properties['frames'][1].gid)
        self.image = self.image_off
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.anim = util.Animation(self, [self.image_off, self.image_on], 0.2, stopped=True)

    def deactivate(self):
        self.image = self.image_off

    def activate(self):
        self.image = self.image_on

    def animate(self):
        self.anim.run(0.5)

    def update(self, dt):
        self.anim.update(dt)


class WireManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_wires, tmx_data, slots):
        pygame.sprite.RenderUpdates.__init__(self)
        self.wire_map = {}
        for tmx_wire in tmx_wires:
            wire_tile = util.get_tile_at_xy(tmx_wire.x, tmx_wire.y)
            slot = next((slot for slot in slots if util.get_tile_at_xy(*slot.pos) == wire_tile), None)
            wire = Wire(tmx_wire, tmx_data, slot)
            self.add(wire)
            if wire_tile in self.wire_map:
                print "Error loading level: Overlapping wires"
                sys.exit()
            self.wire_map[wire_tile] = wire


class Wire(pygame.sprite.Sprite):

    # Activated wires are found in the wire's animation

    # Rects for each part of the wire according to direction (assumes 32x32 wires of width 2)
    DIR_PARTS = {None: pygame.Rect(0,0,32,32), 't': pygame.Rect(15,0,2,16), 'b': pygame.Rect(15,16,2,16),
        'l': pygame.Rect(0,15,16,2), 'r': pygame.Rect(16,15,16,2), 'tl': pygame.Rect(0,0,17,17),
        'tr': pygame.Rect(15,0,17,17), 'bl': pygame.Rect(0,15,17,17), 'br': pygame.Rect(15,15,17,17)}

    def __init__(self, tmx_wire, tmx_data, slot):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [tmx_wire.x, tmx_wire.y]
        self.tilepos = util.get_tile_at_xy(tmx_wire.x, tmx_wire.y)

        self.dirs = tmx_wire.properties['dirs'].split()
        self.image_on = tmx_wire.image
        self.image_off = tmx_data.get_tile_image_by_gid(tmx_wire.properties['frames'][1].gid)
        self.image = self.image_on.copy()
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos

        self.slot = slot
        self.adjacent = {} # to be filled by SourceManager

    def deactivate(self, direction=None):
        self.blit_image_part(self.image_off, direction)

    def activate(self, direction=None):
        self.blit_image_part(self.image_on, direction)

    def set_state(self, state, direction=None):
        if state:
            self.activate(direction)
        else:
            self.deactivate(direction)

    def blit_image_part(self, img, direction=None):
        rect = self.DIR_PARTS[direction]
        self.image.fill((0,0,0,0), rect)
        self.image.blit(img, (rect.x, rect.y), rect)


class SourceManager(pygame.sprite.RenderUpdates):

    def __init__(self, tmx_sources, tmx_data, wires, doors):
        pygame.sprite.RenderUpdates.__init__(self)
        self.source_map = {}
        for tmx_source in tmx_sources:
            source = Source(tmx_source, tmx_data, wires, doors)
            self.add(source)
            source_tile = util.get_tile_at_xy(tmx_source.x, tmx_source.y)
            if source_tile in self.source_map:
                print "Error loading level: Overlapping sources"
                sys.exit()
            self.source_map[source_tile] = source
        self.create_wires_adjacency(wires, doors)

    def create_wires_adjacency(self, all_wires, doors):
        for wire in all_wires:
            for wire_dir in wire.dirs:
                if wire_dir in ['tl', 'tr', 'bl', 'br']:
                    # must end in a door
                    wire.adjacent[wire_dir] = self.get_wire_doors(wire.tilepos, wire_dir, doors)
                elif wire_dir == 'c':
                    # center source
                    if wire.tilepos not in self.source_map:
                        print "Error loading level: Wire ended without source"
                        sys.exit()
                    source = self.source_map[wire.tilepos]
                    wire.adjacent[wire_dir] = source
                    source.adjacent = wire
                else:
                    adj_wire_pos = util.get_tile_from_dir(wire.tilepos[0], wire.tilepos[1], wire_dir)

                    if adj_wire_pos == None:
                        print "Error loading level: Wire with invalid direction or attached to screen border"
                        sys.exit()

                    if adj_wire_pos in all_wires.wire_map:
                        adj_wire = all_wires.wire_map[adj_wire_pos]
                        if util.opposite_dir(wire_dir) in adj_wire.dirs:
                            wire.adjacent[wire_dir] = adj_wire
                            continue

                    # should be at a source
                    if wire.tilepos not in self.source_map:
                        print "Error loading level: Wire ended without source"
                        sys.exit()
                    source = self.source_map[wire.tilepos]
                    wire.adjacent[wire_dir] = source
                    source.adjacent = wire

    def get_wire_doors(self, wire_pos, wire_dir, doors):
        # get all adjacent doors with the appropriate directions
        wire_doors = []
        move_x = util.get_tile_from_dir(wire_pos[0], wire_pos[1], wire_dir[1])
        move_y = util.get_tile_from_dir(wire_pos[0], wire_pos[1], wire_dir[0])
        move_xy = util.get_tile_from_dir(move_x[0], move_x[1], wire_dir[0]) if move_x != None else None

        wire_doors += [door for door in doors.door_map[wire_pos] if door.direction[0] in [wire_dir[0], wire_dir[1]]]
        if move_x:
            wire_doors += [door for door in doors.door_map[move_x] if door.direction[0] in [wire_dir[0], util.opposite_dir(wire_dir[1])]]
        if move_y:
            wire_doors += [door for door in doors.door_map[move_y] if door.direction[0] in [util.opposite_dir(wire_dir[0]), wire_dir[1]]]
        if move_xy:
            wire_doors += [door for door in doors.door_map[move_xy] if door.direction[0] in [util.opposite_dir(wire_dir[0]), util.opposite_dir(wire_dir[1])]]
        return wire_doors

    def update_wires(self):
        doors_state = {}
        for source in self:
            cur_doors_state = source.update_wires()
            for door, state in cur_doors_state.iteritems():
                if state == True:
                    doors_state[door] = True
                else:
                    if door not in doors_state:
                        doors_state[door] = False

        doors_change = False
        for door, state in doors_state.iteritems():
            if state == True:
                if not door.closed:
                    door.close()
                    doors_change = True
            else:
                if door.closed:
                    door.open()
                    doors_change = True
        return doors_change


class Source(pygame.sprite.Sprite):

    def __init__(self, tmx_source, tmx_data, all_wires, doors):
        pygame.sprite.Sprite.__init__(self)
        self.pos = [tmx_source.x, tmx_source.y]

        self.direction = tmx_source.properties['dir']
        self.on = (tmx_source.properties['sourcetype'] == "on")
        self.image = tmx_source.image
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos

        if self.direction == "c":
            self.image_on = tmx_source.image
            self.image_off = tmx_data.get_tile_image_by_gid(tmx_source.properties['frames'][1].gid)
            self.image = self.image_off
            self.on = False

        self.adjacent = None

    def enable(self):
        assert(self.direction == 'c')
        self.image = self.image_on
        self.on = True

    def disable(self):
        assert(self.direction == 'c')
        self.image = self.image_off
        self.on = False

    def update_wires(self):
        doors_state = {} # doors are not closed/opened yet; need to ensure one active wire is enough to close a door
        in_dir = util.opposite_dir(self.direction) if self.direction != 'c' else 'c'
        wires_to_explore = [(self.adjacent, in_dir, self.on)] # (wire, in direction, on state)
        while len(wires_to_explore) > 0:
            wire_or_doors, in_dir, wire_state = wires_to_explore.pop()
            # wire can actually be Wire or list of Doors; shouldn't be Source at this point
            if isinstance(wire_or_doors, Wire):
                wire = wire_or_doors
                if in_dir != 'c': # not from center source
                    wire.set_state(wire_state, in_dir)
                if wire.slot != None and wire.slot.not_gate == True:
                    wire_state = not wire_state
                for out_dir, adj_wire in wire.adjacent.iteritems():
                    if out_dir == in_dir: continue # skip direction we came from
                    wire.set_state(wire_state, out_dir)
                    wires_to_explore.append((adj_wire, util.opposite_dir(out_dir), wire_state))
                
            elif isinstance(wire_or_doors, list):
                for door in wire_or_doors:
                    if wire_state == True:
                        doors_state[door] = True
                    else:
                        if door not in doors_state:
                            doors_state[door] = False
            else:
                print 'Error loading level: Wires with more than one Source'
                sys.exit()
        return doors_state


class TextLayer(pygame.sprite.RenderUpdates):

    def __init__(self, levelnum):
        pygame.sprite.RenderUpdates.__init__(self)
        self.levelnum = levelnum

    def draw_level_title(self):
        sprite = pygame.sprite.Sprite()
        sprite.image, pos = ptext.draw(str(self.levelnum) + ".  " + const.LEVEL_NAMES[self.levelnum], (5, 5), color=(255,255,255),
            fontname=const.FONT_MAIN)
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.add(sprite)

    def draw_lose_text(self):
        sprite = pygame.sprite.Sprite()
        sprite.image, pos = ptext.draw("You cannot let it in! Press R to restart", (const.REAL_WIDTH / 2, const.REAL_HEIGHT - 3),
            anchor=(0.5,1), color=(255,255,255), fontname=const.FONT_MAIN)
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.add(sprite)
